#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

int max(int a, int b) { return a > b ? a : b; }

struct polynom {
  int degree;
  float *coefficients;
};

// Q3.1
struct polynom *create_polynom(int degree, float *coefficients) {
  struct polynom *new_polynom =
      malloc(sizeof(struct polynom));
  new_polynom->coefficients = malloc((degree + 1) * sizeof(float));

  new_polynom->degree = degree;
  for (int i = 0; i <= degree; ++i) {
    new_polynom->coefficients[i] = coefficients[i];
  }

  return new_polynom;
}

// Q3.2
float random_float(float max) {
  return ((float)rand() / (float)RAND_MAX) * max;
}

// Q3.3
struct polynom create_random_polynom(int degree, float max) {
  struct polynom new_polynom;
  new_polynom.coefficients = malloc((degree + 1) * sizeof(float));
  new_polynom.degree = degree;

  for (int i = 0; i <= degree; ++i) {
    new_polynom.coefficients[i] = random_float(max);
  }

  return new_polynom;
}

// Q3.4
struct polynom create_custom_polynom(int degree) {
  struct polynom new_polynom;
  new_polynom.coefficients = malloc((degree + 1) * sizeof(float));
  new_polynom.degree = degree;

  for (int i = 0; i <= degree; ++i) {
    float value;
    int res = scanf("%f", &value);

    if (res != 1) {
      printf("mauvaise valeure, ce coeff est initialisé à 0\n");
      value = 0.0f;
    }

    new_polynom.coefficients[i] = value;
  }

  return new_polynom;
}

// Q3.5
void print_polynom(struct polynom *polynom) {
  for (int i = 0; i <= polynom->degree; ++i) {
    // do not print null coefficient
    if (polynom->coefficients[i] == 0.0f) {
      // on ne fait rien
    } else {
      // printf the correct sign
      if (polynom->coefficients[i] > 0.0f)
        printf("+");

      // print the coefficient
      printf("%.3fX^%d ", polynom->coefficients[i], i);
    }
  }

  printf("\n");
}

// Q3.6
struct polynom sum_polynom(struct polynom p1, struct polynom p2) {
  struct polynom sum;
  sum.degree = max(p1.degree, p2.degree);

  sum.coefficients = malloc((sum.degree + 1) * sizeof(float));

  for (int i = 0; i <= sum.degree; ++i)
    if (i <= p1.degree) {
      if (i <= p2.degree) {
        sum.coefficients[i] = p1.coefficients[i] + p2.coefficients[i];
      } else {
        sum.coefficients[i] = p1.coefficients[i];
      }
    } else {
      sum.coefficients[i] = p2.coefficients[i];
    }
  return sum;
}

// Q3.7
struct polynom prod_polynom(struct polynom p1, struct polynom p2) {
  struct polynom product;
  product.degree = p1.degree + p2.degree;
  product.coefficients = malloc((product.degree + 1) * sizeof(float));

  for (int i = 0; i <= product.degree; ++i) {
    product.coefficients[i] = 0.0f;
    for (int k = 0; k <= i; ++k) {
      product.coefficients[i] += p1.coefficients[k] * p2.coefficients[i - k];
    }
  }

  return product;
}

// Q3.8
float eval_naive_polynom(float x, struct polynom *p) {
  float value = 0.0f;
  for (int i = 0; i <= p->degree; ++i)
    value += p->coefficients[i] * powf(x, i);

  return value;
}

// Q3.9
struct polynom *golden_polynom(int k) {
  struct polynom *list = (struct polynom *)malloc(k * sizeof(struct polynom));

  // creer G0 et stocker dans liste
  float *coeffs_G0 = malloc(sizeof(float) * 1);
  list[0].degree = 0;
  list[0].coefficients = coeffs_G0;
  list[0].coefficients[0] = -1.0f;

  // creer G1 et stocker dans liste
  float *coeffs_G1 = malloc(sizeof(float) * 2);
  list[1].degree = 1;
  list[1].coefficients = coeffs_G1;
  list[1].coefficients[0] = -1.0f;
  list[1].coefficients[1] = 1.0f;

  // creer polynom pour représenter X
  struct polynom PX;
  float *coeffs_PX = malloc(sizeof(float) * 2);
  PX.degree = 1;
  PX.coefficients = coeffs_PX;
  PX.coefficients[0] = 0.0f;
  PX.coefficients[1] = 1.0f;

  // calculer G_2...G_k
  for (int i = 2; i < k; ++i) {
    struct polynom p0 = prod_polynom(PX, list[i - 1]);
    struct polynom p1 = prod_polynom(PX, list[i - 2]);

    list[i] = sum_polynom(p0, p1);

    // on libère les résultats intermédiaires
    free(p0.coefficients);
    free(p1.coefficients);
  }

  // in libère le polynom X
  free(PX.coefficients);

  return list;
}

int main() {
  // on créé et calcule la liste des Golden polynomials
  int N = 6;
  struct polynom *list = golden_polynom(N);

  // on affiche la liste
  for (int i = 0; i < N; ++i) {
    print_polynom(&(list[i]));
  }

  // on libère les coeffs des golden polynom dans la liste
  for (int i = 0; i < N; ++i) {
    free(list[i].coefficients);
  }
  // on libère la liste
  free(list);

  return 0;
}

